#include "cpath.h"
#include "aliases.h"
#include "str8.h"
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>

void join_root_and_leaf(const char *root, const char *leaf, char *dst);

u32 wapp_cpath_join_path(Str8 *dst, const Str8List *parts) {
  if (!dst || !parts) {
    return CPATH_JOIN_INVALID_ARGS;
  }

  if (parts->node_count == 0) {
    return CPATH_JOIN_EMPTY_PARTS;
  }

  Str8 separator = wapp_str8_buf(4);
  wapp_str8_push_back(&separator, PATH_SEP);

  u64 required_capacity = parts->node_count * separator.size + parts->total_size;
  if (dst->capacity < required_capacity) {
    return CPATH_JOIN_INSUFFICIENT_DST_CAPACITY;
  }

  // Handle first node
  const Str8Node *first_node = wapp_str8_list_get(parts, 0);
  wapp_str8_copy_str8_capped(dst, first_node->string);

  // NOTE (Abdelrahman): Uses a while loop instead of a for loop to get rid of
  // MSVC Spectre mitigation warnings
  const Str8Node *node = first_node;
  u64 node_index       = 1;
  bool running         = true;
  while (running && node->next) {
    node = node->next;
    if (node->string->size == 0) {
      continue;
    }

    if (dst->size > 0) {
      char dst_last = wapp_str8_get(dst, dst->size - 1);
      char node_start = wapp_str8_get(node->string, 0);
      bool add_path_sep = dst_last != PATH_SEP && node_start != PATH_SEP;

      if (add_path_sep) {
        wapp_str8_concat_capped(dst, &separator);
      }
    }

    wapp_str8_concat_capped(dst, node->string);

    ++node_index;
    running = node_index < parts->node_count;
  }

  return CPATH_JOIN_SUCCESS;
}

void dirup(char *dst, u64 levels, const char *path) {
  if (levels < 1) {
    return;
  }

  u64 copy_count = 0;
  u64 sep_count = 0;

  u64 full_length;
  u64 length;
  length = full_length = strlen(path);

  if (length > 1 && path[length - 1] == PATH_SEP) {
    --length;
  }

  for (i64 i = length - 1; i >= 0; --i) {
    if (path[i] == PATH_SEP) {
      ++sep_count;
      copy_count = i + 1;

      if (sep_count == levels) {
        break;
      }
    }
  }

  char tmp[256];
  snprintf(tmp, 2, "%c", PATH_SEP);
  // NOTE (Abdelrahman): Conditions stored in variables to silence MSVC warning C5045
  bool insufficient_levels = sep_count < levels;
  bool path_to_copy_is_separator = strncmp(path, tmp, copy_count) != 0;
  if (insufficient_levels && path_to_copy_is_separator) {
    copy_count = 0;
  }

  if (dst == path) {
    memset(&dst[copy_count], 0, full_length - copy_count);
  } else {
    u64 dst_length = strlen(dst);
    memset(dst, 0, dst_length);
    strncpy(dst, path, copy_count);
  }

  u64 final_length = strlen(dst);
  if (final_length > 1) {
    dst[final_length - 1] = '\0';
  } else if (final_length == 0) {
    dst[0] = '.';
  }
}

void join_root_and_leaf(const char *root, const char *leaf, char *dst) {
  u64 root_length = strlen(root);
  u64 leaf_length = strlen(leaf);

  memcpy(dst, root, root_length);

  if (leaf_length == 0) {
    return;
  }

  u64 copy_start = root_length;

  if (root_length > 0 && root[root_length - 1] != PATH_SEP && leaf[0] != PATH_SEP) {
    dst[root_length] = PATH_SEP;
    ++copy_start;
  }

  memcpy(&(dst[copy_start]), leaf, leaf_length);
}
